"use strict";
this.name           = "ils";
this.author         = "Norby";
this.copyright      = "2016 Norby";
this.description    = "Instrument Landing System";
this.licence        = "CC BY-NC-SA 4.0";
this.version        = "1.13"

//internal properties, should not touch
this.$F = null; //FCB pointer
this.$L = null; //last ILS user ship
this.$N = null; //wide NPC in last stage
this.$P = null; //previous heading in FCB
this.$S = null; //the target station
this.$T = null; //Towbar worldScript pointer
this.$Targets = null; //Result of scanner check
this.$Timer = null; //Timer for scanner check
this.$V = null; //visualEffect for the white flashing marker
this.$PI = Math.PI;
this.$maxSpeedRef = 900; // reference for max docking speed: bellow that we try to moderate rotation for visual effects
this.$counter = 0;
this.$debug = false; // turns on tracing messages to the log

//worldscript events
this.startUp = function() {
    this.$T = worldScripts.towbar;
    if( !isValidFrameCallback( this.$F ) )
        this.$F = addFrameCallback( this.$FCB.bind(this) );
        this.$Timer = new Timer(this, this.$Timed.bind(this), 1, 1);//scanner check in every seconds
}

this.equipmentDamaged = this.equipmentDestroyed = function(equipment) {
    if(equipment == "EQ_ILS") this.$ClearILS(); //stop ILS if working
}

/* another oxp cause a bug which turn off ILS immediately in the next lines so do not use this event
this.shipTargetLost = function() {
    log(this.name, "ILS OFF 1");//debug
    this.$ClearILS();
}
*/

this.shipWillDockWithStation = this.shipWillEnterWitchspace = this.shipDied = function() {
    this.$ClearILS(true);//nomsg
}

this.shipWillLaunchFromStation = function(station) {
        if( this.$L && this.$L != player.ship && this.$L.isValid && this.$L.isInSpace ) {
                this.$L.desiredSpeed = 0; //stop landing NPC when player launch
        }
        this.$ClearILS(true);
}

//ILS methods
this.$ClearILS = function(nomsg) {
    this.$N = null;
    this.$P = null;
    this.$S = null;
    if( this.$V && this.$V.isVisualEffect ) {
        this.$V.remove();
        if( !nomsg && this.$L == player.ship )
            player.consoleMessage("ILS off");
    }
    this.$V = null;
    this.$L = null;
//    if( isValidFrameCallback( this.$F ) ) removeFrameCallback( this.$F );
//    log(this.name, "ILS OFF");//debug
}

this.$NPC = function( t ) { //target is a landing NPC ship?
    if( t && t.isPiloted && t.maxSpeed > 0 && t != player.ship
        && ( t.AI == "dockingAI.plist" || t.dockingInstructions )
        && t.target && t.target.isStation && t.target.maxSpeed == 0 ) {
        return(true);
//                 //wide NPC need rolling target station to turn on ILS - disabled since can roll
//                 if( t.target.roll != 0 ) return(true);
//                 var u = t.target.subEntities;
//                 var i = 0;
//                 if( u ) while( u[i] && !u[i].isDock ) i++;
//                 if( u && u[i] && u[i].isDock &&
//                         t.boundingBox.x > Math.min( u[i].boundingBox.x, u[i].boundingBox.y ) ) {
//                 return(true); //not wide so fit into without roll
//                 }
    }
    return(false);
}

this.$FCB = function( delta ) {
//    var w = worldScripts.ils;
//    player.consoleMessage(this.$S+" "+delta); //debug
    var p = player.ship;
    if( !p || !p.isValid ) return;
    // de-referencing for performance
    var L = this.$L; 
    var debug = this.$debug;
    var counter = this.$counter;
    var PI = this.$PI;
    var halfPI = PI / 2;

    var pp = p.position;
    var t = p.target;
    var tis = false; //target is station but not a carrier?
    var npc = false; //there is a landing NPC ship?
    var near = 10000; //NPCs use ILS within this distance
    var noinjdist = 2000; //switch off injectors within this distance
    var maxw = 64; //safe max. width of ship to fit into the dock height without roll

    if( t && ( t.isStation && t.dataKey.indexOf("carrier") == -1 || t == this.$L )
        && !p.missilesOnline //not in missile targeting mode
        && ( p.alertCondition != 3 || !p.weaponsOnline ) //not in red alert with online weapons
        && p.equipmentStatus("EQ_ILS") == "EQUIPMENT_OK" ) { //has ILS
        if( !L || L == p //ILS is not used by NPCs at the moment
            || !L.isValid || !L.isInSpace ) { //or docked/destroyed
            if( t != L ) tis = true;
            else if( L.target && L.target.isStation ) {
                tis = true; //set target to the station when npc is docked
                if( p.target != L.target )
                    t = p.target = L.target; 
            }
        } 
/* dybal: this prevents the player from getting a lock on the station to ask for docking clearance on busy systems; it also causes dozens of messages to the console as the player tries to lcok the station
        else {
            if( p.target != this.$L ) {
                if( pp.distanceTo( this.$L.position ) < p.scannerRange ) {
                    //prevent repeated messages when target is out of range
                    p.target = this.$L;
                }
                player.consoleMessage("ILS locked by "+this.$L.displayName, 5);
            }
        }
*/
    }

    if( player.dockingClearanceStatus === "DOCKING_CLEARANCE_STATUS_GRANTED" 
        || player.dockingClearanceStatus === "DOCKING_CLEARANCE_STATUS_TIMING_OUT" ) {
        if( L && L != p && L.isValid ) 
            L.desiredSpeed = 0; //stop NPC when player has docking clearance
    } else if( L && L != p && L.isValid && L.isInSpace && this.$NPC( L ) ) { 
        //still target to the station? (not if attacked)
        npc = true; //select last npc until is not docked or destroyed
        t = L;
    } else if( !tis ) { //an NPC can use ILS when is not used at the moment
        t = null;
        var targets = this.$Targets; //filled by checkScanner() in timer
        if( targets && targets.length > 0 ) {
            var bestsd = 0; //best speed/distance mean fastest landing so far
            for( var i = 0; i < targets.length; i++ ) {
                var n = targets[i];
                if( this.$NPC( n ) ) {
                    //ILS on the ship which probably arrive first
                    var nd = n.position.distanceTo(n.target);
                    var sd = n.maxSpeed/nd;
                    if( nd < near && nd > noinjdist && n.fuel > 0.2
                        && n.equipmentStatus("EQ_FUEL_INJECTION")
                            == "EQUIPMENT_OK" ) {
                        if( n.injectorSpeedFactor > 0 )
                            sd = n.injectorSpeedFactor * sd;
                        else sd = 7 * sd; //for Oolite 1.80
                    }
                    if( nd < near && ( !t || sd > bestsd ) //first or faster
                        && n.boundingBox.x < 210 ) {//exclude bigships
                        npc = true;
                        t = n; //only one NPC can use ILS at the same time
                        this.$N = null;
                        t.awardEquipment("EQ_ILS");
                        bestsd = sd;
                    } else if( nd > near ) { //no ILS, just a speedup
                        if( n.injectorSpeedFactor > 0 )
                            n.desiredSpeed = n.maxSpeed * n.injectorSpeedFactor;
                        else n.desiredSpeed = 7 * n.maxSpeed;    
                    }
                }
            }
//            if(t) player.consoleMessage(t.name+" "+t.entityPersonality+" ");//debug
        }
    }
    
    if( t && t.isValid && ( tis || npc ) ) {
        if( npc ) { //NPC ILS
            p = t;
            t = t.target;
        }

        var warning = false;
        if( this.$S != t || this.$L != p ) { //first lock on this target
            this.$S = t; //target station
            var m = "";
            if( p != player.ship ) {
                m = " of "+p.name;
//                if( !player.ship.target ) player.ship.target = p; - confusing, removed
            } else warning = true;
            if( !player.ship.docked )
                player.consoleMessage("ILS"+m+" set to "+t.displayName, 5);
//            log(this.name, "ILS set to: "+t);//debug
        }
        this.$L = p; //last ILS user ship
        L = p;
        
        if (debug === true) {
            // the counter is used to limit the tracing messages to a fraction of the FB calls
            if (npc)
                counter = 0;
            else counter++;
            this.$counter = counter;
        }

        var ph = p.heading;
        pp = p.position; //must set again for NPCs
        if( !this.$P ) this.$P = ph;
        var s = this.$S; //station
        var mp = p.maxPitch;
        var mr = p.maxRoll;
        var my = p.maxYaw;
//        log(this.name, "Station: "+s);//debug
        var u = s.subEntities;
        var i = 0;
        if( u ) while( u[i] && !u[i].isDock ) i++;
        var dp = s.position; //determine dock position
        var dr = s.vectorRight; //save for wide npc
        var dz = s.boundingBox.z;
        var dock, dh;

        if( u && u[i] && u[i].isDock ) {
//            var dd = u[i].position.magnitude(); //dock dist. from the center of the station
//            dp = s.position.add( s.heading.direction().multiply(dd) ); //old formula
/* dybal: this fixes part of off-center docking bay problem, the calculated dock position was wrong... each component of this Vector3D constructor is a vector in world-coordinates that should be added to the station position to get the docking bay position
            dp = s.position.add( //better formula for non-centered docks
                Vector3D( s.heading.direction().multiply( dock.position.z ),
                s.vectorUp.direction().multiply( dock.position.y ),
                dr.direction().multiply( dock.position.x ) ) );
*/
            dock = u[i]; // de-referencing to make easier to read
            dp = s.position.add(s.heading.multiply( dock.position.z )).
                            add(s.vectorUp.multiply( dock.position.y )).
                            add(dr.multiply( dock.position.x ));
            var x = dock.boundingBox.x, y = dock.boundingBox.y;
            var doo = dock.orientation; //original dock orientation in station coordinates, normal orientation mean no Torus Station
            var dow = dock.orientation.multiply(s.orientation) // final dock orienatation in world coordinates
            dh = dock.heading.rotateBy(dow); // dock heading in final world coordinates
            if( x < y ) {
                maxw = x; //rotated dock model (coriolis, ico)
                /* dybal: this fixes the dock orientation on tetrahedron and octahedron stations: the dock is taller than its width and not rotated relative to the station, so we have to use the stations vectorUp as the dock's wider direction */
                dr = dow.vectorUp();
            } else {
                maxw = y; //dodo, etc.
                dr = dow.vectorRight();
            }
            dz = u[i].boundingBox.z; //save for wide npc
            if (debug === true && !npc && counter % 60 === 0) // tracing to understand dock orientation
                log(this.name, s.displayName+" ("+s.dataKey+"), Dock x,y:"+x.toFixed(1)+","+y.toFixed(1)+", o_orientation:"+[doo.w,doo.x,doo.y,doo.z]+", w_orientation:"+[dow.w,dow.x,dow.y,dow.z]+", relPosition:"+dock.position+", absPosition:"+dp+", Up:"+dow.vectorUp()+", Right:"+dow.vectorRight()+", Head:"+dh+", Station Up:"+s.vectorUp+", Station Right:"+s.vectorRight+", Head:"+s.heading+", direction:"+dr+", position:"+s.position+", distance:"+pp.distanceTo(dp).toFixed(1))
        }
        var p2d = dp.subtract( pp );
        var dd = p2d.magnitude(); //dock distance, used far below for NPCs also
        if( warning ) {
            if( p.boundingBox.x < maxw )
                player.consoleMessage("You can dock without roll", 5);
            else if( dd < noinjdist )
                player.consoleMessage("Auto roll is active to fit into the dock", 5);
            else player.consoleMessage("Auto roll will be active before the dock", 5);
        }
        var d = dd / 2;
/* dybal: this is the other part of the off-center docking bay fix: if the docking is off-center, as the player gets closer to the dock the vectors to the center of station and to the center of the dock diverge enough that using the station's one would get the playership spiralling around
        var w = s.position.subtract( pp ).direction();
*/
        var w = p2d.direction();
        var an_pd = dh.angleTo(w); // angle from ship position to docking lane (in the lane if angle near PI since dh is vector exiting the dock)
//        player.consoleMessage(d +" "+ s.heading.angleTo(w)); //debug
        if( d < 100 || ( an_pd < halfPI && d < 300 ) ) //fix for rock hermit
            d = -500; //for sure, do not turn back if crossed the point in the dock
        else {
            if( an_pd < 3.14 ) { //player is not before the dock yet
                var na = PI - an_pd;
                na = na * na;
                d += na * 2000;
                //extend distance if apporaching behind from the station
            }
            d = Math.min( 9500, d ); //max. at the Nav Buoy
        }
        var ap = dp.add( dh.multiply( d ) ); //aimed position

        var p2ap_dir = ap.subtract( pp ).direction();
        if( this.$V && this.$V.isVisualEffect ) 
            this.$V.position = ap;
        else 
            this.$V = system.addVisualEffect("ils-marker", ap);
        
        var an_ap = ph.angleTo(p2ap_dir);

        var ato = 0;
        if(ph && this.$P) ato = ph.angleTo(this.$P);//angle to the Previous Heading
        if (debug === true && !npc) // tracing to check heading convergence to center of docking and smoothness of following marker ball
            log(this.name, "counter:"+counter+", distance:"+d.toFixed(1)+", direction:"+w+", heading:"+p.heading+", angle_to_dock:"+an_pd+", aimed:"+ap+", angle_to_ball:"+an_ap+", direction_ball:"+p2ap+", previous_heading:"+this.$P+", angle_to_previous:"+ato);
        /* dybal: changed condition for 'angle' from > 0.0015 to > 0.00015 to get smoother steering */
        if( ato < 0.005 && an_ap > 0.00015
            //steer if no manual steering and not in 0.01 degree
            && p.speed > 1 ) { //and player is not stopped
            
            var reduc = 1; //reduce pitch in the first 0.5s when start steering
            //if(p.script) { //deprecated, startvector based smoothing is better
                        //        if( !p.script.$ILSTimeSinceSteerStart )
                        //                p.script.$ILSTimeSinceSteerStart = delta;
                        //        else p.script.$ILSTimeSinceSteerStart += delta; //count elapsed time
                        //        reduc = 2*Math.max(0.5, p.script.$ILSTimeSinceSteerStart);
                        //}  //reduc start near 0 and reach 1 after 0.5s
            
            //if the above ato value lower then can not start steering in my intel atom netbook
//            player.consoleMessage(this.$P +" vs "+ ph); //debug

            //limit max turn/step and not too accutate
            var r = Math.min(p.maxPitch*reduc*delta, an_ap/12);

            if(p.script ) { //smoother start of ILS steering as the ship has inertia
                var sv = p.script.$ILSSteerStartVect;
                if(sv) 
                    r = Math.min(r, ph.angleTo(sv)/2);
                else {
                    r = 0.1*delta;//need some steering in the first frame
                    p.script.$ILSSteerStartVect = ph; //save steering start vector
                }
            }

            if( !npc && this.$T && this.$T.$TowbarShip
                && this.$T.$TowbarShip.isValid ) {
                var ma = Math.min( 2, p.mass / this.$T.$TowbarShip.mass ); ///small ship max. 2x
                r = r * ma / 3; // 1/3 of the original rate with same mass, min. 1/5 max. 2/3
//                player.consoleMessage("ILS slow steering with towed ship "+ato);//debug
            }
            
            var c = ph.cross(p2ap_dir).direction(); //set the plane where we should rotate in
            var q = p.orientation.rotate( c, -r );
            if (debug === true && !npc) // tracing to check heading convergence to center of docking
                log(this.name, "Steering: counter:"+counter+", distance:"+d.toFixed(1)+", direction:"+w+", heading:"+ph+", angle_to_dock:"+an_pd+", aimed:"+ap+", angle_to_ball:"+an_ap+", direction_ball:"+v+", rotate:"+r);
            p.orientation = q;
        } else if(p.script) 
            p.script.$ILSSteerStartVect = null;
                        //p.script.$ILSTimeSinceSteerStart = 0; //reset delta counter

        this.$P = p.heading;
        ph = p.heading; // steering might have altered it

        if( npc ) {  //NPC ILS
            var ang = 0, dist = 0, n = null;
            var colldist = 2000; //check for collision within this distance
            var sp = p.maxSpeed;
            var coll = false; //check for possible collision with another ship
            var targets = this.$Targets; //filled by checkScanner() in timer
            if( !targets && !(targets.length > 0) ) targets = [];
            targets.push(player.ship);
            for( var i = 0; i < targets.length; i++ ) {
                n = targets[i];
                if( n && n.isValid ) {
                    dist = pp.distanceTo(n);
                    if( dist < colldist ) { //is there a ship within 2km?
                        ang = p.heading.angleTo(pp.subtract(n));
                        //if( n == player.ship ) log(this.name, ang);//debug
                        if( ang > 2 ) { //facing within 65 degree?
                            coll = n; //store the nearest dangerous ship
                            colldist = dist; //used below for desiredSpeed also
                        }
                    }
                }
            }
            if( p.boundingBox.x > maxw && dd < noinjdist ) { //wide ship in the last 2km
                p.destination = ap; //AI will slow down near the ap
                if( dd > p.collisionRadius + 120 + dz/2 ) {
                    p.desiredSpeed = Math.min(sp, 200); //approach slowly
                } else {
                    ang = p.vectorUp.angleTo(dr); //angle between npc and dock
                    var r = 0;
                    if( this.$N || ang > 1.4 && ang < 1.74 ) { //check upside down also
                        //right angle to start docking
                        this.$N = true;//prevent stop after station rolled
                        p.desiredSpeed = sp;
                        p.destination = ap.add(ap.subtract(pp).multiply(1000));
                        r = this.$S.roll * delta; //roll as the station
                    } else {
                        p.desiredSpeed = 0; //wait for the right angle
                        //need roll especially at non rolling stations
                        r = p.maxRoll/2 * delta;
                        if( this.$S.roll > 0 ) r = -r; //roll opposite
                    }
                    p.orientation = p.orientation.rotate(ph, r);
                    //player.consoleMessage(ang+" "+r);//debug
                }
            } else {
                if( dd > noinjdist ) { //use Injector far from the station
                    if( p.injectorSpeedFactor > 0 )
                        p.desiredSpeed = p.injectorSpeedFactor * sp;
                    else p.desiredSpeed = 7 * sp; //for Oolite 1.80
                } else p.desiredSpeed = sp; //normal speed near the station
                //place destination farther to do not slow down when ap is near
                p.destination = ap.add(ap.subtract(pp).multiply(5));
                //this.$V.position = p.destination; //debug
            }
            if( coll ) { //slow down when another ship is near, as slow as near
                p.desiredSpeed = Math.min(p.desiredSpeed, 0.1*colldist);
                //log(this.name, ang);//debug
            }
                p.desiredRange = 0;
//             player.consoleMessage(p.desiredSpeed+" "+Math.round(p.speed));//debug
        } else { //player rotation before the dock if wide ship in the last 2km
            var an2lane = dh.angleTo(p2ap_dir);
            var maxSpeedRef = this.$maxSpeedRef;
            // how fast we have to get in sync... we can do it slowly if the player's ship is going slow
            var abruptness = (p.speed < maxSpeedRef ? Math.max(0.1, 0.9 * p.speed / maxSpeedRef) : 0.9);

            // (no player steering) && (rotation sync necessary) && (start farther if going fast) && (angle to the docking lane <  60 degrees)
            if( ato < 0.005 && p.boundingBox.x > maxw && dd < 10*abruptness*noinjdist && Math.abs(an2lane - PI) < 1) {
                var clock_wise = (s.roll > 0); // if station rotates anti-clockwise, ship facing opposite way should rotate clockwise
                var ang = p.vectorUp.angleTo(dr); //angle between player ship UP and dock RIGHT, should be PI/2 when in sync
                var ang_diff = ang - halfPI;
                var ang_diff_abs = Math.abs(ang_diff);
                var signal = (clock_wise === true ? 1 : -1); // anti-clockwise angles are negative for rotate method
                var max_r = p.maxRoll * delta; // max angle the ship is able to roll in the interval between FCBs
                var r, r_desired;
                var c;
                if( ang_diff_abs < 0.015 ) {
                    // in sync, just roll with the station
                    c = "A";
                    if (s.roll > 0)
                        r_desired = Math.min(s.roll * delta, max_r); //roll as the station
                    else
                        r_desired = - Math.min(-s.roll * delta, max_r); //roll as the station
                    r = r_desired; // just for the debug message bellow
/* dybal: so far it seems the strategies A and C are enough
//                } else if (dd > dock.boundingBox.z+50 ) { // not in the dock yet
//                    c = "B";
//                    r = p.maxRoll/2 * delta;
//                    if( s.roll > 0 ) r = -r; //roll opposite gets there faster
*/
                } else { 
                    c = "C";
                    if (ang_diff > 0 && clock_wise === true || ang_diff < 0 && clock_wise === false) {
                        // lagging, should rotate the same direction as the station
                        r_desired = s.roll*delta + signal * abruptness * ang_diff_abs;
                    } else {
                        // advanced, should rotate opposite direction as the station
                        r_desired = s.roll*delta - signal * abruptness * ang_diff_abs;
                    }
                    // limit the angle to the ship roll capability
                    if (r_desired > 0)
                        r =  Math.min(r_desired, max_r);
                    else
                        r = - Math.min(-r_desired, max_r);
                }
                if (debug === true && !npc) // tracing to check heading convergence to center of docking
                    log(this.name, "counter:"+counter+", distance:"+dd.toFixed(0)+"("+d.toFixed(1)+"), angle_to_lane:"+an2lane.toFixed(5)+", angle_to_dock:"+ang.toFixed(3)+"("+ang_diff.toFixed(5)+"), clock-wise:"+clock_wise+", signal:"+signal+", Case:"+c+", delta:"+delta.toFixed(4)+", rotate:"+r.toFixed(5)+", r_desired:"+r_desired.toFixed(4)+", max_r:"+max_r.toFixed(4)+", abrupt:"+abruptness.toFixed(2));
                p.orientation = p.orientation.rotate(ph, r);
                //player.consoleMessage(ang+" "+r);//debug
            }
        }
    } else this.$ClearILS(); //no target or a new target is not station, etc.
}

this.$Timed = function() {
    var c = player.ship; //check around player
    if( this.$S ) c = this.$S; //check around last station if set
    if( !c || !c.isValid ) return;
    this.$Targets = c.checkScanner(true);//much faster than filteredEntities
}
